@@ -26,7 +26,7 @@
  *
  *  Hardware cursor support added by Emmanuel Marty (core@ggi-project.org)
  *  Smart redraw scrolling, arbitrary font width support, 512char font support
- *  and software scrollback added by
+ *  and software scrollback added by 
  *                         Jakub Jelinek (jj@ultra.linux.cz)
  *
  *  Random hacking by Martin Mares <mj@ucw.cz>
@@ -111,7 +111,7 @@ static int softback_lines;
 /* console mappings */
 static int first_fb_vc;
 static int last_fb_vc = MAX_NR_CONSOLES - 1;
-static int fbcon_is_default = 1;
+static int fbcon_is_default = 1; 
 static int fbcon_has_exited;
 static int primary_device = -1;
 static int fbcon_has_console_bind;
@@ -405,7 +405,7 @@ static void cursor_timer_handler(unsigned long dev_addr)
 	struct fb_info *info = (struct fb_info *) dev_addr;
 	struct fbcon_ops *ops = info->fbcon_par;
 
-	schedule_work(&info->queue);
+	queue_work(system_power_efficient_wq, &info->queue);
 	mod_timer(&ops->cursor_timer, jiffies + HZ/5);
 }
 
@@ -450,8 +450,8 @@ static int __init fb_console_setup(char *this_opt)
 
 	while ((options = strsep(&this_opt, ",")) != NULL) {
 		if (!strncmp(options, "font:", 5))
-			strcpy(fontname, options + 5);
-
+			strlcpy(fontname, options + 5, sizeof(fontname));
+		
 		if (!strncmp(options, "scrollback:", 11)) {
 			options += 11;
 			if (*options) {
@@ -466,7 +466,7 @@ static int __init fb_console_setup(char *this_opt)
 			} else
 				return 1;
 		}
-
+		
 		if (!strncmp(options, "map:", 4)) {
 			options += 4;
 			if (*options) {
@@ -491,8 +491,8 @@ static int __init fb_console_setup(char *this_opt)
 				first_fb_vc = 0;
 			if (*options++ == '-')
 				last_fb_vc = simple_strtoul(options, &options, 10) - 1;
-			fbcon_is_default = 0;
-		}
+			fbcon_is_default = 0; 
+		}	
 
 		if (!strncmp(options, "rotate:", 7)) {
 			options += 7;
@@ -552,34 +552,6 @@ static int do_fbcon_takeover(int show_logo)
 		info_idx = -1;
 	} else {
 		fbcon_has_console_bind = 1;
-	}
-
-	return err;
-}
-
-static int fbcon_takeover(int show_logo)
-{
-	int err, i;
-
-	if (!num_registered_fb)
-		return -ENODEV;
-
-	if (!show_logo)
-		logo_shown = FBCON_LOGO_DONTSHOW;
-
-	for (i = first_fb_vc; i <= last_fb_vc; i++)
-		con2fb_map[i] = info_idx;
-
-	err = take_over_console(&fb_con, first_fb_vc, last_fb_vc,
-				fbcon_is_default);
-
-	if (err) {
-		for (i = first_fb_vc; i <= last_fb_vc; i++) {
-			con2fb_map[i] = -1;
-		}
-		info_idx = -1;
-	} else {
-		fbcon_has_console_bind = 1;
 #ifdef CONFIG_FB_CON_DECOR
 		fbcon_decor_init();
 #endif
@@ -905,7 +877,7 @@ static int set_con2fb_map(int unit, int newidx, int user)
 /*
  *  Low Level Operations
  */
-/* NOTE: fbcon cannot be __init: it may be called from take_over_console later */
+/* NOTE: fbcon cannot be __init: it may be called from do_take_over_console later */
 static int var_to_display(struct display *disp,
 			  struct fb_var_screeninfo *var,
 			  struct fb_info *info)
@@ -973,7 +945,7 @@ static const char *fbcon_startup(void)
 	info = registered_fb[info_idx];
 	if (!info)
 		return NULL;
-
+	
 	owner = info->fbops->owner;
 	if (!try_module_get(owner))
 		return NULL;
@@ -1287,13 +1259,21 @@ static void fbcon_clear(struct vc_data *vc, int sy, int sx, int height,
 	if (!height || !width)
 		return;
 
-	if (sy < vc->vc_top && vc->vc_top == logo_lines)
+	if (sy < vc->vc_top && vc->vc_top == logo_lines) {
 		vc->vc_top = 0;
+		/*
+		 * If the font dimensions are not an integral of the display
+		 * dimensions then the ops->clear below won't end up clearing
+		 * the margins.  Call clear_margins here in case the logo
+		 * bitmap stretched into the margin area.
+		 */
+		fbcon_clear_margins(vc, 0);
+	}
 
- 	if (fbcon_decor_active(info, vc)) {
- 		fbcon_decor_clear(vc, info, sy, sx, height, width);
- 		return;
- 	}
+	if (fbcon_decor_active(info, vc)) {
+		fbcon_decor_clear(vc, info, sy, sx, height, width);
+		return;
+	}
 
 	/* Split blits that cross physical y_wrap boundary */
 
@@ -1320,8 +1300,8 @@ static void fbcon_putcs(struct vc_data *vc, const unsigned short *s,
 			fbcon_decor_putcs(vc, info, s, count, ypos, xpos);
 		else
 			ops->putcs(vc, info, s, count, real_y(p, ypos), xpos,
-				   get_color(vc, info, scr_readw(s), 1),
-				   get_color(vc, info, scr_readw(s), 0));
+			get_color(vc, info, scr_readw(s), 1),
+			get_color(vc, info, scr_readw(s), 0));
 	}
 }
 
@@ -1337,14 +1317,6 @@ static void fbcon_clear_margins(struct vc_data *vc, int bottom_only)
 {
 	struct fb_info *info = registered_fb[con2fb_map[vc->vc_num]];
 	struct fbcon_ops *ops = info->fbcon_par;
-
-	if (!fbcon_is_inactive(vc, info)) {
-	 	if (fbcon_decor_active(info, vc)) {
-	 		fbcon_decor_clear_margins(vc, info, bottom_only);
- 		} else {
-			ops->clear_margins(vc, info, bottom_only);
-		}
-	}
 }
 
 static void fbcon_cursor(struct vc_data *vc, int mode)
@@ -1454,7 +1426,7 @@ static __inline__ void ywrap_up(struct vc_data *vc, int count)
 	struct fb_info *info = registered_fb[con2fb_map[vc->vc_num]];
 	struct fbcon_ops *ops = info->fbcon_par;
 	struct display *p = &fb_display[vc->vc_num];
-
+	
 	p->yscroll += count;
 	if (p->yscroll >= p->vrows)	/* Deal with wrap */
 		p->yscroll -= p->vrows;
@@ -1473,7 +1445,7 @@ static __inline__ void ywrap_down(struct vc_data *vc, int count)
 	struct fb_info *info = registered_fb[con2fb_map[vc->vc_num]];
 	struct fbcon_ops *ops = info->fbcon_par;
 	struct display *p = &fb_display[vc->vc_num];
-
+	
 	p->yscroll -= count;
 	if (p->yscroll < 0)	/* Deal with wrap */
 		p->yscroll += p->vrows;
@@ -1540,7 +1512,7 @@ static __inline__ void ypan_down(struct vc_data *vc, int count)
 	struct fb_info *info = registered_fb[con2fb_map[vc->vc_num]];
 	struct display *p = &fb_display[vc->vc_num];
 	struct fbcon_ops *ops = info->fbcon_par;
-
+	
 	p->yscroll -= count;
 	if (p->yscroll < 0) {
 		ops->bmove(vc, info, 0, 0, p->vrows - vc->vc_rows,
@@ -2051,7 +2023,7 @@ static void fbcon_bmove(struct vc_data *vc, int sy, int sx, int dy, int dx,
 {
 	struct fb_info *info = registered_fb[con2fb_map[vc->vc_num]];
 	struct display *p = &fb_display[vc->vc_num];
-
+	
 	if (fbcon_is_inactive(vc, info))
 		return;
 
@@ -2069,7 +2041,7 @@ static void fbcon_bmove(struct vc_data *vc, int sy, int sx, int dy, int dx,
 			p->vrows - p->yscroll);
 }
 
-static void fbcon_bmove_rec(struct vc_data *vc, struct display *p, int sy, int sx,
+static void fbcon_bmove_rec(struct vc_data *vc, struct display *p, int sy, int sx, 
 			    int dy, int dx, int height, int width, u_int y_break)
 {
 	struct fb_info *info = registered_fb[con2fb_map[vc->vc_num]];
@@ -2109,10 +2081,10 @@ static void fbcon_bmove_rec(struct vc_data *vc, struct display *p, int sy, int s
 	}
 
 	if (fbcon_decor_active(info, vc) && sy == dy && height == 1) {
- 		/* must use slower redraw bmove to keep background pic intact */
- 		fbcon_decor_bmove_redraw(vc, info, sy, sx, dx, width);
- 		return;
- 	}
+		/* must use slower redraw bmove to keep background pic intact */
+		fbcon_decor_bmove_redraw(vc, info, sy, sx, dx, width);
+		return;
+	}
 
 	ops->bmove(vc, info, real_y(p, sy), sx, real_y(p, dy), dx,
 		   height, width);
@@ -2165,7 +2137,7 @@ static void updatescrollmode(struct display *p,
 	}
 }
 
-static int fbcon_resize(struct vc_data *vc, unsigned int width,
+static int fbcon_resize(struct vc_data *vc, unsigned int width, 
 			unsigned int height, unsigned int user)
 {
 	struct fb_info *info = registered_fb[con2fb_map[vc->vc_num]];
@@ -2221,20 +2193,21 @@ static int fbcon_switch(struct vc_data *vc)
 
 	info = registered_fb[con2fb_map[vc->vc_num]];
 	ops = info->fbcon_par;
+
 	prev_console = ops->currcon;
 	if (prev_console != -1)
 		old_info = registered_fb[con2fb_map[prev_console]];
 
 #ifdef CONFIG_FB_CON_DECOR
-	if (!fbcon_decor_active_vc(vc) && info->fix.visual == FB_VISUAL_DIRECTCOLOR) {
-		struct vc_data *vc_curr = vc_cons[prev_console].d;
-		if (vc_curr && fbcon_decor_active_vc(vc_curr)) {
-			/* Clear the screen to avoid displaying funky colors during
-			 * palette updates. */
-			memset((u8*)info->screen_base + info->fix.line_length * info->var.yoffset,
-			       0, info->var.yres * info->fix.line_length);
+		if (!fbcon_decor_active_vc(vc) && info->fix.visual == FB_VISUAL_DIRECTCOLOR) {
+			struct vc_data *vc_curr = vc_cons[prev_console].d;
+			if (vc_curr && fbcon_decor_active_vc(vc_curr)) {
+				/* Clear the screen to avoid displaying funky colors during
+				 * * palette updates. */
+				memset((u8*)info->screen_base + info->fix.line_length * info->var.yoffset,
+					0, info->var.yres * info->fix.line_length);
+			}
 		}
-	}
 #endif
 
 	if (softback_top) {
@@ -2297,15 +2270,15 @@ static int fbcon_switch(struct vc_data *vc)
 			fbcon_del_cursor_timer(old_info);
 	}
 
-	if (fbcon_decor_active_vc(vc)) {
+ 	if (fbcon_decor_active_vc(vc)) {
 		struct vc_data *vc_curr = vc_cons[prev_console].d;
 
 		if (!vc_curr->vc_decor.theme ||
 			strcmp(vc->vc_decor.theme, vc_curr->vc_decor.theme) ||
 			(fbcon_decor_active_nores(info, vc_curr) &&
-			 !fbcon_decor_active(info, vc_curr))) {
-			fbcon_decor_disable(vc, 0);
-			fbcon_decor_call_helper("modechange", vc->vc_num);
+			!fbcon_decor_active(info, vc_curr))) {
+				fbcon_decor_disable(vc, 0);
+				fbcon_decor_call_helper("modechange", vc->vc_num);
 		}
 	}
 
@@ -2357,10 +2330,11 @@ static int fbcon_switch(struct vc_data *vc)
 	    ops->update_start(info);
 	}
 
-	fbcon_set_palette(vc, color_table);
+	fbcon_set_palette(vc, color_table); 	
 	fbcon_clear_margins(vc, 0);
 
 	if (logo_shown == FBCON_LOGO_DRAW) {
+
 		logo_shown = fg_console;
 		/* This is protected above by initmem_freed */
 		fb_show_logo(info, ops->rotate);
@@ -2416,13 +2390,12 @@ static int fbcon_blank(struct vc_data *vc, int blank, int mode_switch)
 		}
 	}
 
-	if (!fbcon_is_inactive(vc, info)) {
+ 	if (!fbcon_is_inactive(vc, info)) {
 		if (ops->blank_state != blank) {
 			ops->blank_state = blank;
 			fbcon_cursor(vc, blank ? CM_ERASE : CM_DRAW);
 			ops->cursor_flash = (!blank);
-
-			if (!(info->flags & FBINFO_MISC_USEREVENT)) {
+            if (!(info->flags & FBINFO_MISC_USEREVENT)) {
 				if (fb_blank(info, blank)) {
 					if (fbcon_decor_active(info, vc))
 						fbcon_decor_blank(vc, info, blank);
@@ -2551,7 +2524,7 @@ static int fbcon_do_set_font(struct vc_data *vc, int w, int h,
 			vc->vc_complement_mask >>= 1;
 			vc->vc_s_complement_mask >>= 1;
 		}
-
+			
 		/* ++Edmund: reorder the attribute bits */
 		if (vc->vc_can_do_color) {
 			unsigned short *cp =
@@ -2574,7 +2547,7 @@ static int fbcon_do_set_font(struct vc_data *vc, int w, int h,
 			vc->vc_complement_mask <<= 1;
 			vc->vc_s_complement_mask <<= 1;
 		}
-
+			
 		/* ++Edmund: reorder the attribute bits */
 		{
 			unsigned short *cp =
@@ -2615,11 +2588,10 @@ static int fbcon_do_set_font(struct vc_data *vc, int w, int h,
 			cols = vc->vc_decor.twidth;
 			rows = vc->vc_decor.theight;
 		}
+
 		cols /= w;
 		rows /= h;
-
 		vc_resize(vc, cols, rows);
-
 		if (CON_IS_VISIBLE(vc) && softback_buf)
 			fbcon_update_softback(vc);
 	} else if (CON_IS_VISIBLE(vc)
@@ -2703,7 +2675,7 @@ static int fbcon_set_font(struct vc_data *vc, struct console_font *font, unsigne
 	/* Check if the same font is on some other console already */
 	for (i = first_fb_vc; i <= last_fb_vc; i++) {
 		struct vc_data *tmp = vc_cons[i].d;
-
+		
 		if (fb_display[i].userfont &&
 		    fb_display[i].fontdata &&
 		    FNTSUM(fb_display[i].fontdata) == csum &&
@@ -2750,9 +2722,9 @@ static int fbcon_set_palette(struct vc_data *vc, unsigned char *table)
 
 	if (fbcon_is_inactive(vc, info)
 #ifdef CONFIG_FB_CON_DECOR
-			|| vc->vc_num != fg_console
+		|| vc->vc_num != fg_console
 #endif
-		)
+	)
 		return -EINVAL;
 
 	if (!CON_IS_VISIBLE(vc))
@@ -2779,55 +2751,55 @@ static int fbcon_set_palette(struct vc_data *vc, unsigned char *table)
 		fb_copy_cmap(fb_default_cmap(1 << depth), &palette_cmap);
 
 	if (fbcon_decor_active(info, vc_cons[fg_console].d) &&
-	    info->fix.visual == FB_VISUAL_DIRECTCOLOR) {
+       info->fix.visual == FB_VISUAL_DIRECTCOLOR) {
 
-		u16 *red, *green, *blue;
-		int minlen = min(min(info->var.red.length, info->var.green.length),
-				     info->var.blue.length);
-		int h;
+       u16 *red, *green, *blue;
+       int minlen = min(min(info->var.red.length, info->var.green.length),
+                    info->var.blue.length);
+       int h;
 
-		struct fb_cmap cmap = {
-			.start = 0,
-			.len = (1 << minlen),
-			.red = NULL,
-			.green = NULL,
-			.blue = NULL,
-			.transp = NULL
-		};
+       struct fb_cmap cmap = {
+           .start = 0,
+           .len = (1 << minlen),
+           .red = NULL,
+           .green = NULL,
+           .blue = NULL,
+           .transp = NULL
+       };
 
-		red = kmalloc(256 * sizeof(u16) * 3, GFP_KERNEL);
+       red = kmalloc(256 * sizeof(u16) * 3, GFP_KERNEL);
 
-		if (!red)
-			goto out;
+       if (!red)
+           goto out;
 
-		green = red + 256;
-		blue = green + 256;
-		cmap.red = red;
-		cmap.green = green;
-		cmap.blue = blue;
+       green = red + 256;
+       blue = green + 256;
+       cmap.red = red;
+       cmap.green = green;
+       cmap.blue = blue;
 
-		for (i = 0; i < cmap.len; i++) {
-			red[i] = green[i] = blue[i] = (0xffff * i)/(cmap.len-1);
-		}
+       for (i = 0; i < cmap.len; i++) {
+           red[i] = green[i] = blue[i] = (0xffff * i)/(cmap.len-1);
+       }
 
-		h = fb_set_cmap(&cmap, info);
-		fbcon_decor_fix_pseudo_pal(info, vc_cons[fg_console].d);
-		kfree(red);
+       h = fb_set_cmap(&cmap, info);
+       fbcon_decor_fix_pseudo_pal(info, vc_cons[fg_console].d);
+       kfree(red);
 
-		return h;
+       return h;
 
-	} else if (fbcon_decor_active(info, vc_cons[fg_console].d) &&
-		   info->var.bits_per_pixel == 8 && info->bgdecor.cmap.red != NULL)
-		fb_set_cmap(&info->bgdecor.cmap, info);
+   } else if (fbcon_decor_active(info, vc_cons[fg_console].d) &&
+          info->var.bits_per_pixel == 8 && info->bgdecor.cmap.red != NULL)
+       fb_set_cmap(&info->bgdecor.cmap, info);
 
-out:	return fb_set_cmap(&palette_cmap, info);
+out:   return fb_set_cmap(&palette_cmap, info);
 }
 
 static u16 *fbcon_screen_pos(struct vc_data *vc, int offset)
 {
 	unsigned long p;
 	int line;
-
+	
 	if (vc->vc_num != fg_console || !softback_lines)
 		return (u16 *) (vc->vc_origin + offset);
 	line = offset / vc->vc_size_row;
@@ -3046,13 +3018,12 @@ static void fbcon_modechanged(struct fb_info *info)
 		rows = FBCON_SWAP(ops->rotate, info->var.yres, info->var.xres);
 		cols /= vc->vc_font.width;
 		rows /= vc->vc_font.height;
-
-		if (!fbcon_decor_active_nores(info, vc)) {
-			vc_resize(vc, cols, rows);
+ 		if (!fbcon_decor_active_nores(info, vc)) {
+           vc_resize(vc, cols, rows);
 		} else {
-			fbcon_decor_disable(vc, 0);
-			fbcon_decor_call_helper("modechange", vc->vc_num);
-		}
+        	fbcon_decor_disable(vc, 0);
+           	fbcon_decor_call_helper("modechange", vc->vc_num);
+       	}
 
 		updatescrollmode(p, info, vc);
 		scrollback_max = 0;
@@ -3099,8 +3070,9 @@ static void fbcon_set_all_vcs(struct fb_info *info)
 		cols /= vc->vc_font.width;
 		rows /= vc->vc_font.height;
 		if (!fbcon_decor_active_nores(info, vc)) {
-			vc_resize(vc, cols, rows);
+        	vc_resize(vc, cols, rows);
 		}
+
 	}
 
 	if (fg != -1)
@@ -3661,8 +3633,9 @@ static void fbcon_start(void)
 			}
 		}
 
+		do_fbcon_takeover(0);
 		console_unlock();
-		fbcon_takeover(0);
+
 	}
 }
 
@@ -3767,9 +3740,9 @@ static void __exit fb_console_exit(void)
 	fbcon_deinit_device();
 	device_destroy(fb_class, MKDEV(0, 0));
 	fbcon_exit();
+	do_unregister_con_driver(&fb_con);
 	console_unlock();
-	unregister_con_driver(&fb_con);
-}
+}	
 
 module_exit(fb_console_exit);
 
